import itertools
import os
import sys
import importlib
import logbook
from importlib.machinery import ModuleSpec, SourceFileLoader

_logger = logbook.Logger(__name__)


class NoInitFileFound(Exception):
    pass


def import_file(filename):
    """Given a path to a file, imports it as a Python module
    """
    module_name = _setup_module_name_for_import(filename)

    return _import_using_new_importlib(module_name, filename)


def set_package_name(directory, name):
    """Given a directory, establishes that directory as a pseudo-package, enabling clearer future imports
    """
    directory = _normalize_path(directory)
    if directory not in _cached_package_names:
        _create_package_module(name, directory)
        _cached_package_names[directory] = name


def _import_using_new_importlib(module_name, filename):
    is_package = os.path.isdir(filename) or filename.endswith('__init__.py')
    if os.path.isdir(filename):
        filename = os.path.join(filename, "__init__.py")

    package_name = module_name if is_package else module_name.rsplit('.', 1)[0]
    if package_name != module_name and package_name not in sys.modules:
        # need to import the package first
        pkg = SourceFileLoader(package_name, os.path.join(
            os.path.dirname(filename), '__init__.py')).load_module()
        sys.modules[package_name] = pkg

    return __import__(module_name, fromlist=[''])

_package_name_generator = ('_{0}'.format(x) for x in itertools.count())


def _generate_package_name():
    for suggested in _package_name_generator:
        if not _package_name_exists(suggested):
            return suggested


def _package_name_exists(pkg_name):
    return pkg_name in sys.modules


def _setup_module_name_for_import(filename):
    return _create_new_module_name(filename)

_cached_package_names = {}


def _create_new_module_name(filename):
    _logger.trace("Creating new package for {0}", filename)
    nonpackage_dir, remainder = _split_nonpackage_dir(filename)
    _logger.trace("After split: {0}, {1}", nonpackage_dir, remainder)
    package_name = _cached_package_names.get(nonpackage_dir, None)
    if package_name is None:
        package_name = _generate_package_name()
        sys.modules[package_name] = _create_package_module(
            package_name, nonpackage_dir)
        _cached_package_names[nonpackage_dir] = package_name
    returned = '{0}.{1}'.format(package_name, remainder)
    if returned.endswith('.__init__'):
        returned = returned.rsplit('.', 1)[0]
    return returned


def _split_nonpackage_dir(path):
    if not os.path.isdir(path):
        nonpackage_dir, module = os.path.split(_normalize_path(path))
        module = _make_module_name(module).split(".")
    else:
        nonpackage_dir = path
        module = []
    while os.path.isfile(os.path.join(nonpackage_dir, "__init__.py")):
        if '.' in os.path.split(nonpackage_dir)[-1]:
            # we cannot import from such packages, stop traversing upwards...
            break
        nonpackage_dir, current_component = os.path.split(nonpackage_dir)
        module.insert(0, current_component)
        _logger.trace("Now at {0}, {1}", nonpackage_dir, module)
    if not module:
        raise NoInitFileFound(
            "Could not find __init__.py file in {0}".format(path))
    return nonpackage_dir, ".".join(module)


def _normalize_path(path):
    return os.path.normpath(os.path.abspath(str(path)))


def _make_module_name(filename):
    assert filename.endswith('.py') or filename.endswith('.pyc')
    return filename.rsplit(".", 1)[0].replace(os.path.sep, ".")


def _create_package_module(name, path):
    spec = ModuleSpec(origin=path, name=name, loader=SourceFileLoader(name, path), is_package=True)
    spec.submodule_search_locations.append(path)
    returned = importlib.util.module_from_spec(spec)
    returned.__path__ = [path]
    sys.modules[name] = returned
    return returned



import os
import sys

import emport
import pytest

@pytest.mark.skipif(sys.version_info < (3, 6), reason='requires Python 3.6')
def test_module_specs(tmpdir, recwarn):

    package_dir = tmpdir.join('package')

    subpackage_dir = package_dir.join('sub')

    for p in (package_dir, subpackage_dir):
        with p.join('__init__.py').open('w', ensure=True) as f:
            pass

    with subpackage_dir.join('utils.py').open('w') as f:
        pass

    filename = subpackage_dir.join('module.py')
    with filename.open('w', ensure=True) as f:
        f.write('from . import utils')

    mod = emport.import_file(str(filename))

    for m in (mod, mod.utils):

        metapackage_name, remainder = m.__spec__.name.split('.', 1)
        assert remainder == 'package.sub.{}'.format(m.__name__.split('.')[-1]) # pylint: disable=no-member
        assert mod.__spec__.parent == '{0}.package.sub'.format(metapackage_name)      # pylint: disable=no-member


    metapackage = sys.modules[metapackage_name]


    assert metapackage.__spec__ is not None
    assert metapackage.__spec__.origin == str(package_dir.dirname)
    assert metapackage.__spec__.name == metapackage_name
    # TODO: restore this
    #assert metapackage.__spec__.parent == metapackage_name
    assert metapackage.__spec__.submodule_search_locations is not None
    assert str(package_dir.dirname) in metapackage.__spec__.submodule_search_locations
    assert metapackage.__package__ == metapackage_name

    assert not recwarn.list



import os
import emport

import pytest


def test_import_from_non_package(tmpdir, module_file_factory):
    _import_and_verify(
        [module_file_factory(tmpdir, filename='{0}.py'.format(i)) for i in range(5)])


def test_multiple_imports_from_same_non_package(tmpdir, module_file_factory):
    file1 = module_file_factory(tmpdir)
    file2 = module_file_factory(tmpdir)
    [module1] = _import_and_verify([file1])
    [module2] = _import_and_verify([file2])
    assert _PACKAGE_NAME(module1) == _PACKAGE_NAME(module2)
    assert module1 is not module2


def test_importing_sub_package(tmpdir, module_file_factory):
    subpackage_dir = tmpdir.join('subpackage')
    subpackage_dir.join('__init__.py').ensure(file=True)
    full_filename, expected_value = module_file_factory(
        subpackage_dir, "module.py")
    module = emport.import_file(full_filename)
    assert module.__name__.endswith('.subpackage.module')


def test_importing_sub_package_and_subdir(tmpdir, module_file_factory):
    to_verify = module_file_factory(tmpdir)
    [module] = _import_and_verify([to_verify])
    package_name = _PACKAGE_NAME(module)

    subdir = tmpdir.join('sub')

    subdir.join('__init__.py').ensure(file=True)
    [sub_module] = _import_and_verify([module_file_factory(subdir)])
    assert (_PACKAGE_NAME(module) + '.sub') == _PACKAGE_NAME(sub_module)


def test_importing_different_directories_same_escaping(tmpdir, module_file_factory):
    dir1 = tmpdir.join("pkg+")
    dir2 = tmpdir.join("pkg-")
    filenames_and_expected_values = [module_file_factory(x) for x in [dir1, dir2]]
    module1, module2 = _import_and_verify(filenames_and_expected_values)
    assert _PACKAGE_NAME(module1) != _PACKAGE_NAME(module2)


def test_importing_dotted_name(tmpdir, module_file_factory):
    path_components = ["a", "b.c", "d"]
    path = tmpdir
    for c in path_components:
        path = path.join(c)
        path.join('__init__.py').ensure(file=True)
    filename, expected_value = module_file_factory(path, "module.py")
    _import_and_verify([(filename, expected_value)])


def test_importing_directory_no_init_file(tmpdir):
    with pytest.raises(emport.NoInitFileFound):
        emport.import_file(str(tmpdir))


@pytest.mark.parametrize('init_py_directly', [True, False])
def test_importing_directory(init_py_directly, tmpdir, module_file_factory):
    directory = str(tmpdir.join("pkg"))
    os.makedirs(directory)
    _, expected_value = module_file_factory(directory, "__init__.py")
    if init_py_directly:
        module = emport.import_file(os.path.join(directory, "__init__.py"))
    else:
        module = emport.import_file(directory)

    assert module.__name__.endswith(".pkg")
    assert module.value == expected_value


def test_importing_subdirectory_init_file(tmpdir, module_file_factory):
    directory = tmpdir.join('pkg')
    module_file_factory(directory, '__init__.py')
    subdir = directory.join('subpkg')
    filename, expected_value = module_file_factory(subdir, '__init__.py')
    module = emport.import_file(filename)
    assert module.value == expected_value


def _import_and_verify(filenames_and_expected_values):
    returned = []
    for filename, expected_value in filenames_and_expected_values:
        module = emport.import_file(filename)
        assert module.value == expected_value
        returned.append(module)
    return returned


def _PACKAGE_NAME(m):
    return m.__name__.rsplit(".", 1)[0]



import pytest
from emport import import_file


def test_relative_imports(directory):
    module = import_file(str(directory.join('proxy.py')))
    assert module.file_1_value == 'file_1'


@pytest.fixture
def directory(tmpdir):
    returned = tmpdir.join('dir')

    returned.join('__init__.py').ensure(file=True)


    with returned.join('file_1.py').open('w') as f:
        f.write('value = "file_1"')

    with returned.join('proxy.py').open('w') as f:
        f.write('from .file_1 import value as file_1_value')

    return returned



from uuid import uuid4

import emport


def test_import_set_name(tmpdir):
    package_name = '__pkg_{0}__'.format(str(uuid4()).replace('-', '_'))
    emport.set_package_name(tmpdir, package_name)

    with tmpdir.join('file.py').open('w') as f:
        f.write('value = 666')

    mod = emport.import_file(f.name)
    assert mod.__name__ == '{0}.file'.format(package_name)



import pytest
import warnings
from uuid import uuid4
from emport import import_file


@pytest.mark.parametrize('add_init_py', [True, False])
def test_importing_doesnt_emit_warnings(tmpdir, recwarn, add_init_py):
    value = str(uuid4())
    warnings.simplefilter('always')

    directory = tmpdir.join('files')

    filename = directory.join('testfile.py')
    with directory.join('utils.py').open('w', ensure=True) as f:
        f.write('value = {0!r}'.format(value))

    with filename.open('w') as f:
        f.write('from .utils import value as new_value')

    if add_init_py:
        with directory.join('__init__.py').open('w') as f:
            pass

    module = import_file(str(filename))
    assert module.new_value == value
    assert len(recwarn.list) == 0
